/* Merge AppleDouble Header and DataFork Files */
/* into an AppleSingle File                    */
/*                  Copyright(c)2000 by T.Toda */
#include <stdio.h>

#ifdef	MSDOS
#define	READMODE	"rb"
#define	WRITEMODE	"wb"
#else
#ifdef	WIN32
#define	READMODE	"rb"
#define	WRITEMODE	"wb"
#else
#define	READMODE	"r"
#define	WRITEMODE	"w"
#endif
#endif

static	void	fputhex(unsigned long, int, FILE *);
static	unsigned long	makelong(unsigned char *);
static	unsigned short	makeshort(unsigned char *);
static	void	divlong(unsigned char *, unsigned long);
static	void	divshort(unsigned char *, unsigned short);

int	main(int argc, char **argv, char **envp)
{
	FILE	*fphead, *fpdata, *fpout;
	struct{
		unsigned char	magic[4];
		unsigned char	version[4];
		char	filler[16];
		unsigned char	entries[2];
	}	header;
	struct{
		unsigned char	id[4];
		unsigned char	offset[4];
		unsigned char	length[4];
	}	entry;
	unsigned long	i_magic, i_offset, i_length;
	unsigned short	i_entries;
	unsigned short	i;
	unsigned long	maxend, byteread, datalength;
	int	c;

	if( argc > 4 || argc < 2 ){
		fputs("Usage : ", stderr);
		fputs(argv[0], stderr);
		fputs(" datafile [headerfile [outfile]]\n", stderr);
		return	4;
	}
	if(( fpdata = fopen(argv[1],READMODE)) == NULL ){
		fputs("Can't Open '", stderr);
		fputs(argv[1], stderr);
		fputs("'.\n", stderr);
		return	8;
	}
	if( argc < 3 ){
		fphead = stdin;
	}
	else if(( fphead = fopen(argv[2],READMODE)) == NULL ){
		fputs("Can't Open '", stderr);
		fputs(argv[2], stderr);
		fputs("'.\n", stderr);
		return	8;
	}
	if( argc < 4 ){
		fpout = stdout;
	}
	else if(( fpout = fopen(argv[3],WRITEMODE)) == NULL ){
		fputs("Can't Open '",stderr);
		fputs(argv[3],stderr);
		fputs("'.\n",stderr);
		return	8;
	}

	fseek(fpdata, 0L, 2);
	datalength = ftell(fpdata);
	rewind(fpdata);

	fread(&header, sizeof(header), 1, fphead);
	maxend = byteread = sizeof(header);
	i_entries = makeshort(header.entries);
	i_magic = makelong(header.magic);
	if( i_magic != 0x00051607 ){
		fputs("Invalid AppleDouble Identifier : ", stderr);
		fputhex(i_magic, 8, stderr);
		putc('\n', stderr);
		return	8;
	}
	header.magic[3] = 0x00;
	divshort(header.entries, i_entries + 1);
	fwrite(&header, sizeof(header), 1, fpout);

	for( i = 0 ; i < i_entries ; i++ ){
		fread(&entry, sizeof(entry), 1, fphead);
		byteread += sizeof(entry);
		i_offset = makelong(entry.offset);
		i_length = makelong(entry.length);
		i_offset += sizeof(entry);
		divlong(entry.offset, i_offset);
		if( maxend < i_offset + i_length ){
			maxend = i_offset + i_length;
		}
		fwrite(&entry, sizeof(entry), 1, fpout);

	}
	divlong(entry.id, 1);
	divlong(entry.offset, maxend);
	divlong(entry.length, datalength);
	fwrite(&entry, sizeof(entry), 1, fpout);
	byteread += sizeof(entry);

	while( byteread < maxend ){
		if( (c = getc(fphead)) == EOF ){
			fputs("Error: Header File Shorter than Estimated\n", stderr);
			return	12;
		}
		putc(c, fpout);
		byteread++;
	}
	if( getc(fphead) != EOF ){
		fputs("Warning: Header File Longer than Estimated\n", stderr);
	}

	for( byteread = 0 ; byteread < datalength ; byteread++ ){
		if( (c = getc(fpdata)) == EOF ){
			fputs("Error: Data File Shorter than Pre-Seeked\n", stderr);
			return	12;
		}
		putc(c, fpout);
	}
	if( getc(fpdata) != EOF ){
		fputs("Warning: Data File Longer than Pre-Seeked\n", stderr);
	}
	return	0;
}

static void	fputhex(unsigned long c, int l, FILE *fp)
{
	static	char	const	hexnum[] = "0123456789ABCDEF";
	char	result[9];

	if(l > 8)l = 8;
	if(l < 1)l = 1;
	result[l] = 0;
	result[--l] = hexnum[c % 0x10];
	while( l-- )
		result[l] = hexnum[(c >>= 4) % 0x10];
	fputs(result, fp);
}

static	unsigned long	makelong(unsigned char *buffer)
{
	return	( ((long)buffer[0] << 24 ) | 
	      	  ((long)buffer[1] << 16 ) | 
	      	  ((long)buffer[2] <<  8 ) | 
	      	  ((long)buffer[3] ) );
}

static	unsigned short	makeshort(unsigned char *buffer)
{
	return	( ((long)buffer[0] <<  8 ) | 
	      	  ((long)buffer[1] ) );
}

static	void	divlong(unsigned char *buffer, unsigned long data)
{
	buffer[0] = (unsigned char)(( data >> 24 ) & 0x000000ff );
	buffer[1] = (unsigned char)(( data >> 16 ) & 0x000000ff );
	buffer[2] = (unsigned char)(( data >>  8 ) & 0x000000ff );
	buffer[3] = (unsigned char)(  data         & 0x000000ff );
}

static	void	divshort(unsigned char *buffer, unsigned short data)
{
	buffer[0] = (unsigned char)(( data >>  8 ) & 0x000000ff );
	buffer[1] = (unsigned char)(  data         & 0x000000ff );
}
