/* Decode BinHex 4.0 into AppleDouble */
/*         Copyright(c)2003 by T.Toda */
#include <stdio.h>
#include <string.h>

#ifdef	MSDOS
#define	WRITEMODE	"wb"
#else
#ifdef	WIN32
#define	WRITEMODE	"wb"
#else
#define	WRITEMODE	"w"
#endif
#endif

#define	LINEBUF	80

static	void	fputhex(unsigned long, int, FILE *);
static	unsigned long	makelong(unsigned char *);
static	unsigned short	makeshort(unsigned char *);
static	void	divlong(unsigned char *, unsigned long);
static	void	divshort(unsigned char *, unsigned short);
static	int	get6bit(void), get8bit(void), runlength(void);
static	void	calc_crc(unsigned char);
static	FILE	*fpin;

static	unsigned short	crc_now = 0;

int	main(int argc, char **argv, char **envp)
{
	FILE	*fphead, *fpdata;
	struct{
		unsigned char	magic[4];
		unsigned char	version[4];
		char	filler[16];
		unsigned char	entries[2];
	}	appleheader;
	struct{
		unsigned char	id[4];
		unsigned char	offset[4];
		unsigned char	length[4];
	}	appleentry;
	struct{
		char	filetype[4];
		char	creator[4];
		unsigned char	finderflag[2];
		unsigned char	vertical[2];
		unsigned char	horizontal[2];
		unsigned char	folderid[2];
		unsigned char	extended[16];
	}	finderinfo;
	char	linebuf[LINEBUF];
	unsigned long	lastoffset, rf_len, df_len;
	int	c;
	unsigned short	i, fnlen, crc0;
	unsigned char	numbuf[4];

	if( argc > 4 || argc < 3 ){
		fputs("Usage : ", stderr);
		fputs(argv[0], stderr);
		fputs(" infile headerfile [datafile]\n", stderr);
		return	4;
	}
	if( (fpin = fopen(argv[1], "r")) == NULL ){
		fputs("Can't Open '", stderr);
		fputs(argv[1], stderr);
		fputs("'.\n", stderr);
		return	8;
	}
	if( (fphead = fopen(argv[2], "r")) != NULL ){
		fputs("Destinate File '", stderr);
		fputs(argv[2], stderr);
		fputs("' Exists. Overwrite? (Y/N)", stderr);
		c = getchar();
		if( (c != 'Y') && (c != 'y') ){
			return	8;
		}
		while( getchar() != '\n' );
	}
	fclose(fphead);
	if( (fphead = fopen(argv[2], WRITEMODE)) == NULL ){
		fputs("Can't Open '", stderr);
		fputs(argv[2], stderr);
		fputs("'.\n", stderr);
		return	8;
	}
	if( argc < 4 ){
		fpdata = stdout;
	}
	else{
		if( (fpdata = fopen(argv[3], "r")) != NULL ){
			fputs("Destinate File '", stderr);
			fputs(argv[3], stderr);
			fputs("' Exists. Overwrite? (Y/N)", stderr);
			c = getchar();
			if( (c != 'Y') && (c != 'y') ){
				return	8;
			}
			while( getchar() != '\n' );
		}
		fclose(fpdata);
		if( (fpdata = fopen(argv[3], WRITEMODE)) == NULL ){
			fputs("Can't Open '",stderr);
			fputs(argv[3],stderr);
			fputs("'.\n",stderr);
			return	8;
		}
	}
	for(;;){
		if( fgets(linebuf, LINEBUF, fpin) == NULL ){
			return	4;
		}
		if( strcmp("(This file must be converted with BinHex 4.0)\n", linebuf) == 0 ){
			break;
		}
	}
	while( (c = getc(fpin)) != ':' ){
		if( c == EOF ){
			return	4;
		}
	}

	memset(&appleheader, 0, sizeof(appleheader));
	lastoffset = sizeof(appleheader);
	lastoffset += sizeof(appleentry) * 3;
	divshort(appleheader.entries, 3);
		/* 3 = Name, Finder, RF (no Date, Comment and DF) */
	divlong(appleheader.magic, 0x00051607);
	divlong(appleheader.version, 0x00020000);
	fwrite(&appleheader, sizeof(appleheader), 1, fphead);

	/* 3: Header for Real Name */
	fnlen = (unsigned char)runlength();
	if( fnlen > LINEBUF ){
		fputs("Too Long File Name\n", stderr);
		fnlen = LINEBUF;
	}
	divlong(appleentry.id, 3L);
	divlong(appleentry.length, (unsigned long)fnlen);
	divlong(appleentry.offset, lastoffset);
	lastoffset += fnlen;
	fwrite(&appleentry, sizeof(appleentry), 1, fphead);

	/* Store the Body for Real Name */
	for( i = 0 ; i < fnlen ; i++ ){
		 linebuf[i] = runlength();
	}
	if( c = runlength() ){
		fputs("Invalid File Name Terminator: ", stderr);
		fputhex(c, 2, stderr);
		putc('\n', stderr);
	}

	/* Store the Body for Finder Info */
	memset(&finderinfo, 0, sizeof(finderinfo));
	for( i = 0 ; i < 4 ; i++ ){
		finderinfo.filetype[i] = runlength();
	}
	for( i = 0 ; i < 4 ; i++ ){
		finderinfo.creator[i] = runlength();
	}
	for( i = 0 ; i < 2 ; i++ ){
		finderinfo.finderflag[i] = runlength();
	}

	/* 9: Header for Finder Info */
	divlong(appleentry.id, 9L);
	divlong(appleentry.length, 32L);
	divlong(appleentry.offset, lastoffset);
	lastoffset += 32L;
	fwrite(&appleentry, sizeof(appleentry), 1, fphead);

	/* 2: Header for Resource Fork */
	for( i = 0 ; i < 4 ; i++ ){
		numbuf[i] = runlength();
	}
	df_len = makelong(numbuf);
	for( i = 0 ; i < 4 ; i++ ){
		numbuf[i] = runlength();
	}
	rf_len = makelong(numbuf);
	divlong(appleentry.id, 2L);
	memcpy(appleentry.length, numbuf, 4);
	divlong(appleentry.offset, lastoffset);
	lastoffset += rf_len;
	fwrite(&appleentry, sizeof(appleentry), 1, fphead);

	/* Check the CRC */
	calc_crc(0);
	calc_crc(0);
	crc0 = crc_now;
	for( i = 0 ; i < 2 ; i++ ){
		numbuf[i] = c = runlength();
		if( c == EOF ){
			fputs("Error: End of File in the Header\n", stderr);
			return	12;
		}
	}
	if( makeshort(numbuf) != crc0 ){
		fputs("CRC for Header Mismatch: Calculated = ", stderr);
		fputhex((unsigned long)crc0, 4, stderr);
		fputs(", Recorded = ", stderr);
		fputhex((unsigned long)makeshort(numbuf), 4, stderr);
		putc('\n', stderr);
	}
	crc_now = 0;

	/* 3: Body of Real Name */
	fwrite(linebuf, fnlen, 1, fphead);

	/* 9: Body of Finder Info */
	fwrite(&finderinfo, sizeof(finderinfo), 1, fphead);

	/* 1: Body of Data Fork */
	while( df_len-- ){
		if( (c = runlength()) == EOF ){
			fputs("Error: End of File in Estimated Data Fork\n", stderr);
			return	12;
		}
		putc(c, fpdata);
	}

	/* Check the CRC */
	calc_crc(0);
	calc_crc(0);
	crc0 = crc_now;
	for( i = 0 ; i < 2 ; i++ ){
		numbuf[i] = runlength();
	}
	if( makeshort(numbuf) != crc0 ){
		fputs("CRC for Data Fork Mismatch: Calculated = ", stderr);
		fputhex((unsigned long)crc0, 4, stderr);
		fputs(", Recorded = ", stderr);
		fputhex((unsigned long)makeshort(numbuf), 4, stderr);
		putc('\n', stderr);
	}
	crc_now = 0;

	/* 2: Body of Resource Fork */
	while( rf_len-- ){
		if( (c = runlength()) == EOF ){
			fputs("Error: End of File in Estimated Resource Fork\n", stderr);
			return	12;
		}
		putc(c, fphead);
	}

	/* Check the CRC */
	calc_crc(0);
	calc_crc(0);
	crc0 = crc_now;
	for( i = 0 ; i < 2 ; i++ ){
		numbuf[i] = runlength();
	}
	if( makeshort(numbuf) != crc0 ){
		fputs("CRC for Resource Fork Mismatch: Calculated = ", stderr);
		fputhex((unsigned long)crc0, 4, stderr);
		fputs(", Recorded = ", stderr);
		fputhex((unsigned long)makeshort(numbuf), 4, stderr);
		putc('\n', stderr);
	}
	crc_now = 0;

	if( runlength() != EOF ){
		fputs("Warning: Extra Data on the Tail\n", stderr);
	}

	return	0;
}

static	int	runlength()
{
	static	int	c0 = 0, wait = 0, cont = 0;
	int	c;

	for(;;){
		if( c0 == EOF ){
			return	EOF;
		}
		else if( wait ){
			wait--;
		}
		else if( (c = get8bit()) != 0x90 ){
			c0 = c;
			cont = 0;
		}
		/* kf[^̒0x90Ƃ̋͋Kił͞B     */
		/* kƂ͊Ř􂳂ɐf[^ƊŘ􂷂Ƃ߂ƁA     */
		/* sƓp^[ēxWJƂ߂\łA */
		/* ̔f͌҂̏ꍇ if( (c = get8bit()) != 0x90 )   */
		/* O҂̏ꍇ if( ((c = get8bit()) != 0x90) || cont )       */
		else{
			wait = get8bit();
			if( wait == EOF ){
				c0 = EOF;
			}
			else if( wait == 0 ){
				c0 = 0x90;
				cont = 0;
			}
			else{
				if( cont )wait++;
				cont = 1;
				if( wait == 1 ){
					wait = 0;
					continue;
				}
				else{
					wait -= 2;
				}
			}
		}
		calc_crc((unsigned char)c0);
		return	c0;
	}
}

#define	CRC16_POLYNOMIAL	0x1021;
static	void	calc_crc(unsigned char c)
{
	unsigned short	bit;

	for( bit = 0 ; bit < 8 ; bit++ ){
		if( crc_now & 0x8000 ){
			crc_now <<= 1;
			crc_now ^= CRC16_POLYNOMIAL;
		}
		else{
			crc_now <<= 1;
		}
		if( c & 0x80 ){
			crc_now ^= 1;
		}
		c <<= 1;
	}
}

static	int	get8bit()
{
	static	int	i = 0;
	static	int	buffer[3];
	int	c;

	switch( i ){
	default:
		if( (c = get6bit()) == EOF ){
			buffer[0] = buffer[1] = buffer[2] = EOF;
		}
		else{
			buffer[0] = c << 2;
			if( (c = get6bit()) == EOF ){
				buffer[0] = buffer[1] = buffer[2] = EOF;
			}
			else{
				buffer[0] |= c >> 4;
				buffer[0] &= 0xff;
				buffer[1] = c << 4;
				if( (c = get6bit()) == EOF ){
					buffer[1] = buffer[2] = EOF;
				}
				else{
					buffer[1] |= c >> 2;
					buffer[1] &= 0xff;
					buffer[2] = c << 6;
					if( (c = get6bit()) == EOF ){
						buffer[2] = EOF;
					}
					else{
						buffer[2] |= c;
						buffer[2] &= 0xff;
					}
				}
			}
		}
		i = 1;
		return	buffer[0];
	case	1:
		if( buffer[1] != EOF )i = 2;
		return	buffer[1];
	case	2:
		if( buffer[2] != EOF )i = 0;
		return	buffer[2];
	}
}

static	int	get6bit()
{
	int	c;

	for(;;){
		c = getc(fpin);
		if( c >= '!' && c <= '-' ){
			return	c - '!';
		}
		else if( c >= '0' && c <= '6' ){
			return	c + 13 - '0';
		}
		else if( c >= '8' && c <= '9' ){
			return	c + 20 - '8';
		}
		else if( c >= '@' && c <= 'N' ){
			return	c + 22 - '@';
		}
		else if( c >= 'P' && c <= 'V' ){
			return	c + 37 - 'P';
		}
		else if( c >= 'X' && c <= '[' ){
			return	c + 44 - 'X';
		}
		else if( c >= '`' && c <= 'f' ){
			return	c + 48 - '`';
		}
		else if( c >= 'h' && c <= 'm' ){
			return	c + 55 - 'h';
		}
		else if( c >= 'p' && c <= 'r' ){
			return	c + 61 - 'p';
		}
		else if( c == ':' || c == EOF ){
			return	EOF;
		}
	}
}

static void	fputhex(unsigned long c, int l, FILE *fp)
{
	static	char	const	hexnum[] = "0123456789ABCDEF";
	char	result[9];

	if(l > 8)l = 8;
	if(l < 1)l = 1;
	result[l] = 0;
	result[--l] = hexnum[c % 0x10];
	while( l-- )
		result[l] = hexnum[(c >>= 4) % 0x10];
	fputs(result, fp);
}

static	unsigned long	makelong(unsigned char *buffer)
{
	return	( ((long)buffer[0] << 24 ) | 
	      	  ((long)buffer[1] << 16 ) | 
	      	  ((long)buffer[2] <<  8 ) | 
	      	  ((long)buffer[3] ) );
}

static	unsigned short	makeshort(unsigned char *buffer)
{
	return	( ((long)buffer[0] <<  8 ) | 
	      	  ((long)buffer[1] ) );
}

static	void	divlong(unsigned char *buffer, unsigned long data)
{
	buffer[0] = (unsigned char)(( data >> 24 ) & 0x000000ff );
	buffer[1] = (unsigned char)(( data >> 16 ) & 0x000000ff );
	buffer[2] = (unsigned char)(( data >>  8 ) & 0x000000ff );
	buffer[3] = (unsigned char)(  data         & 0x000000ff );
}

static	void	divshort(unsigned char *buffer, unsigned short data)
{
	buffer[0] = (unsigned char)(( data >>  8 ) & 0x000000ff );
	buffer[1] = (unsigned char)(  data         & 0x000000ff );
}
