/* Encode BinHex 4.0 from AppleDouble */
/*         Copyright(c)2003 by T.Toda */
#include <stdio.h>
#include <string.h>
#include <memory.h>

#ifdef	MSDOS
#define	READMODE	"rb"
#else
#ifdef	WIN32
#define	READMODE	"rb"
#else
#define	READMODE	"r"
#endif
#endif

#define	LINELEN	64

static	void	fputhex(unsigned long, int, FILE *);
static	unsigned long	makelong(unsigned char *);
static	unsigned short	makeshort(unsigned char *);
static	void	divlong(unsigned char *, unsigned long);
static	void	divshort(unsigned char *, unsigned short);
static	void	put6bit(int), end6bit(void), line6bit(void);
static	void	runlength(int), endrun(void);
static	void	calc_crc(unsigned char);
static	FILE	*fpout;

static	char	hqxchar[] 
	= "!\"#$%&'()*+,-012345689@ABCDEFGHIJKLMNPQRSTUVXYZ[`abcdefhijklmpqr";
static	char	*idtext[]
	= {"Unknown", "Data Fork", "Resource Fork", "Real Name", 
	   "Comment", "BW Icon", "Color Icon", "File Info", 
	   "File Dates", "Finder Info", "Mac File Info", "ProDos File Info", 
	   "MSDOS File Info", "Short Name", "AFP File Info", "Directory ID"};
static	unsigned char	hqx_buf[3] = {0, 0, 0};
static	unsigned short	crc_now = 0;
static	int	bufptr = 0, run_data = EOF, run_count = 0;

int	main(int argc, char **argv, char **envp)
{
	FILE	*fphead, *fpdata;
	struct{
		unsigned char	magic[4];
		unsigned char	version[4];
		char	filler[16];
		unsigned char	entries[2];
	}	appleheader;
	struct{
		unsigned char	id[4];
		unsigned char	offset[4];
		unsigned char	length[4];
	}	appleentry;
	struct{
		char	filetype[4];
		char	creator[4];
		unsigned char	finderflag[2];
		unsigned char	vertical[2];
		unsigned char	horizontal[2];
		unsigned char	folderid[2];
		unsigned char	extended[16];
	}	finderinfo;
	unsigned long	i_magic, i_id;
	unsigned long	i_offset[16], i_length[16];
	unsigned long	df_len, rf_len;
	unsigned char	numbuf[4];
	unsigned short	i_entries, i;
	int	c;

	if( argc > 4 || argc < 2 ){
		fputs("Usage : ", stderr);
		fputs(argv[0], stderr);
		fputs(" datafile [headerfile [outfile]]\n", stderr);
		return	4;
	}
	if(( fpdata = fopen(argv[1],READMODE)) == NULL ){
		fputs("Can't Open '", stderr);
		fputs(argv[1], stderr);
		fputs("'.\n", stderr);
		return	8;
	}
	if( argc < 3 ){
		fphead = stdin;
	}
	else if(( fphead = fopen(argv[2],READMODE)) == NULL ){
		fputs("Can't Open '", stderr);
		fputs(argv[2], stderr);
		fputs("'.\n", stderr);
		return	8;
	}
	if( argc < 4 ){
		fpout = stdout;
	}
	else if(( fpout = fopen(argv[3],"w")) == NULL ){
		fputs("Can't Open '",stderr);
		fputs(argv[3],stderr);
		fputs("'.\n",stderr);
		return	8;
	}

	fputs("(This file must be converted with BinHex 4.0)\n\n:", fpout);
	line6bit();

	fseek(fpdata, 0L, 2);
	df_len = ftell(fpdata);
	rewind(fpdata);

	fread(&appleheader, sizeof(appleheader), 1, fphead);
	i_entries = makeshort(appleheader.entries);
	i_magic = makelong(appleheader.magic);
	if( i_magic != 0x00051607 ){
		fputs("Invalid AppleDouble Identifier : ", stderr);
		fputhex(i_magic, 8, stderr);
		putc('\n', stderr);
		return	8;
	}

	for( i = 0 ; i < 16 ; i++ ){
		i_offset[i] = 0;
	}
	for( i = 0 ; i < i_entries ; i++ ){
		fread(&appleentry, sizeof(appleentry), 1, fphead);
		i_id = makelong(appleentry.id);
		if( i_id > 16 ){
			fputs("Warning: Unknown Entry ID", stderr);
			fputhex(i_id, 8, stderr);
			putc('\n', stderr);
		}
		else if( i_offset[i_id] ){
			fputs("Warning: Duplicated Entry ID", stderr);
			fputhex(i_id, 1, stderr);
			putc('(', stderr);
			fputs(idtext[i_id], stderr);
			fputs(")\n", stderr);
		}
		else{
			i_offset[i_id] = makelong(appleentry.offset);
			i_length[i_id] = makelong(appleentry.length);
		}
	}

	/* 3: Real Name */
	if( i_offset[3] ){
		fseek(fphead, (signed long)i_offset[3], 0);
		if( i_length[3] > 63 )i_length[3] = 63;
		runlength((unsigned char)i_length[3]);
		for( i = 0 ; i < (unsigned short)i_length[3] ; i++ ){
			runlength(getc(fphead));
		}
	}
	else{
		runlength(0);
	}
	runlength(0);

	/* 9: Finder Info */
	if( i_offset[9] ){
		fseek(fphead, (signed long)i_offset[9], 0);
		fread(&finderinfo, sizeof(finderinfo), 1, fphead);
	}
	else{
		memset(&finderinfo, 0, sizeof(finderinfo));
	}
	for( i = 0 ; i < 4 ; i++ ){
		runlength(finderinfo.filetype[i]);
	}
	for( i = 0 ; i < 4 ; i++ ){
		runlength(finderinfo.creator[i]);
	}
	for( i = 0 ; i < 2 ; i++ ){
		runlength(finderinfo.finderflag[i]);
	}

	/* 1: Header for Data Fork */
	divlong(numbuf, df_len);
	for( i = 0 ; i < 4 ; i++ ){
		runlength(numbuf[i]);
	}

	/* 2: Header for Resource Fork */
	rf_len = i_offset[2] ? i_length[2] : 0;
	divlong(numbuf, rf_len);
	for( i = 0 ; i < 4 ; i++ ){
		runlength(numbuf[i]);
	}

	/* Calculate CRC */
	calc_crc(0);
	calc_crc(0);
	divshort(numbuf, crc_now);
	for( i = 0 ; i < 2 ; i++ ){
		runlength(numbuf[i]);
	}
	crc_now = 0;

	/* 1: Body of Data Fork */
	while( df_len-- ){
		if( (c = getc(fpdata)) == EOF ){
			fputs("Error: Data File Shorter than Pre-Seeked\n", stderr);
			endrun();
			end6bit();
			return	12;
		}
		runlength(c);
	}
	if( getc(fpdata) != EOF ){
		fputs("Warning: Data File Longer than Pre-Seeked\n", stderr);
	}

	/* Calculate CRC */
	calc_crc(0);
	calc_crc(0);
	divshort(numbuf, crc_now);
	for( i = 0 ; i < 2 ; i++ ){
		runlength(numbuf[i]);
	}
	crc_now = 0;

	/* 2: Body of Resource Fork */
	if( i_offset[2] ){
		fseek(fphead, (signed long)i_offset[2], 0);
		while( rf_len-- ){
			if( (c = getc(fphead)) == EOF ){
				fputs("Error: End of File in Estimated Resource Fork\n", stderr);
				endrun();
				end6bit();
				return	12;
			}
			runlength(c);
		}
	}

	/* Calculate CRC */
	calc_crc(0);
	calc_crc(0);
	divshort(numbuf, crc_now);
	for( i = 0 ; i < 2 ; i++ ){
		runlength(numbuf[i]);
	}
	crc_now = 0;

	endrun();
	end6bit();
	return	0;
}

static	void	runlength(int c)
{
	c &= 0xff;
	calc_crc(c);

	if( (c == run_data) && (run_count < 0xff) ){
		run_count++;
	}
	else{
		endrun();
		put6bit(c);
		if( c == 0x90 ){
			put6bit(0);
		}
		c = run_data;
	}
}

static	void	endrun()
{
	if( run_count ){
		put6bit(0x90);
		put6bit(run_count);
		run_count = 0;
	}
}

#define	CRC16_POLYNOMIAL	0x1021;
static	void	calc_crc(unsigned char c)
{
	unsigned short	bit;

	for( bit = 0 ; bit < 8 ; bit++ ){
		if( crc_now & 0x8000 ){
			crc_now <<= 1;
			crc_now ^= CRC16_POLYNOMIAL;
		}
		else{
			crc_now <<= 1;
		}
		if( c & 0x80 ){
			crc_now ^= 1;
		}
		c <<= 1;
	}
}

static	void	put6bit(int c)
{
	int	i;

	hqx_buf[bufptr++] = (unsigned char)c;
	if( bufptr >= 3 ){
		putc(hqxchar[ ( hqx_buf[0] >> 2 ) & 0x3f ], fpout);
		line6bit();
		putc(hqxchar[ (( hqx_buf[0] << 4 ) 
		             | ( hqx_buf[1] >> 4 )) & 0x3f ], fpout);
		line6bit();
		putc(hqxchar[ (( hqx_buf[1] << 2 ) 
		             | ( hqx_buf[2] >> 6 )) & 0x3f ], fpout);
		line6bit();
		putc(hqxchar[ hqx_buf[2] & 0x3f ], fpout);
		line6bit();
		for( i = 0 ; i < 3 ; i++ ){
			hqx_buf[i] = 0;
		}
		bufptr = 0;
	}
}

static	void	end6bit()
{
	if( bufptr > 0 ){
		putc(hqxchar[  ( hqx_buf[0] >> 2 )  & 0x3f ], fpout);
		line6bit();
		putc(hqxchar[ (( hqx_buf[0] << 4 ) 
		             | ( hqx_buf[1] >> 4 )) & 0x3f ], fpout);
		line6bit();
		if( bufptr > 1 ){
			putc(hqxchar[ (( hqx_buf[1] << 2 ) 
			             | ( hqx_buf[2] >> 6 )) & 0x3f ], fpout);
			line6bit();
		}
	}
	fputs(":\n", fpout);
}

static	void	line6bit()
{
	static	len = 0;

	if( ++len >= LINELEN ){
		putc('\n', fpout);
		len = 0;
	}
}

static void	fputhex(unsigned long c, int l, FILE *fp)
{
	static	char	const	hexnum[] = "0123456789ABCDEF";
	char	result[9];

	if(l > 8)l = 8;
	if(l < 1)l = 1;
	result[l] = 0;
	result[--l] = hexnum[c % 0x10];
	while( l-- )
		result[l] = hexnum[(c >>= 4) % 0x10];
	fputs(result, fp);
}

static	unsigned long	makelong(unsigned char *buffer)
{
	return	( ((long)buffer[0] << 24 ) | 
	      	  ((long)buffer[1] << 16 ) | 
	      	  ((long)buffer[2] <<  8 ) | 
	      	  ((long)buffer[3] ) );
}

static	unsigned short	makeshort(unsigned char *buffer)
{
	return	( ((long)buffer[0] <<  8 ) | 
	      	  ((long)buffer[1] ) );
}

static	void	divlong(unsigned char *buffer, unsigned long data)
{
	buffer[0] = (unsigned char)(( data >> 24 ) & 0x000000ff );
	buffer[1] = (unsigned char)(( data >> 16 ) & 0x000000ff );
	buffer[2] = (unsigned char)(( data >>  8 ) & 0x000000ff );
	buffer[3] = (unsigned char)(  data         & 0x000000ff );
}

static	void	divshort(unsigned char *buffer, unsigned short data)
{
	buffer[0] = (unsigned char)(( data >>  8 ) & 0x000000ff );
	buffer[1] = (unsigned char)(  data         & 0x000000ff );
}
